# Ninklings configuration
import configparser, os
configFileDirOptions = [
    '.',
    os.path.expanduser('~'),
    '/etc'
]

# Default configuration
config = configparser.ConfigParser()
config['ninklings'] = {
    'datadir': '{}/../data'.format(os.path.dirname(__file__)), #TODO change to subdir ninklings when making public
    'projectdir': '/tmp'
}

# Read configuration file
for configFileDir in configFileDirOptions:
    configFile = os.path.join(configFileDir,'ninklings.cfg')
    if os.path.exists(configFile):
        config.read(configFile)
        break #only reads the first config file found



# -*- coding: utf-8 -*-
"""Simulations.

Random simulations for testing.
"""

import random, numpy as np, pandas as pd
import scipy.special as sps
from scipy import stats

class SimScores:
    """
    TODO refactor -> SimScores should be interface class
    Make a similuted scores object.
    Produces the scores to use, and
    methods for calculating false positives/negatives

    >>> SimScores.empirical_param_calc(testdata.logFC,testdata['adj.P.Val']<=.05)
    """
    def __init__(
            self, allGenes, trueGenes, distro='gamma', onlyPositiveScores = True,
            gamma_shape_t = 1.4, gamma_scale_t = .5, gamma_loc_t = .15, gamma_shape_f = 1, gamma_scale_f = .5,
            unif_low_t = 0, unif_high_t = .05, unif_low_f = 0, unif_high_f = 1
    ):
        """
        Sets up simulated scores
        distro choices are: gamma, uniform

        TODO ?include gaussian for t-stat score?
        """
        self.genes = pd.Series(allGenes)
        self.truth = self.genes.isin(trueGenes)
        self.shape = (sum(self.truth),len(self.genes)-sum(self.truth))
        self.data = pd.DataFrame({'gene': self.genes, 'truth': self.truth})
        self.distro = distro
        if self.distro == 'gamma':
            self.distro_params = {
                't': (gamma_shape_t, gamma_scale_t, gamma_loc_t),
                'f': (gamma_shape_f, gamma_scale_f, 0)
            }
            self.data.loc[self.truth,'score'] = np.random.gamma(gamma_shape_t, gamma_scale_t, self.shape[0]) + gamma_loc_t
            self.data.loc[~self.truth,'score'] = np.random.gamma(gamma_shape_f, gamma_scale_f, self.shape[1])
        elif self.distro == 'uniform':
            self.distro_params = {
                't': (unif_low_t, unif_high_t),
                'f': (unif_low_f, unif_high_f)
            }
            self.data.loc[self.truth,'score'] = np.random.uniform(self.distro_params['t'][0], self.distro_params['t'][1], self.shape[0])
            self.data.loc[~self.truth,'score'] = np.random.uniform(self.distro_params['f'][0], self.distro_params['f'][1], self.shape[1])
        else:
            raise Exception('%s distro not supported' % distro)
        if not onlyPositiveScores: raise NotImplementedError

    def plot_scores(self,density=False):
        """Plot scores
        
        Visually screen the similated scores.
        """
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        ax_y2 = ax.twinx()
        if self.distro == 'gamma':
            for torf in ('f','t'):
                gamma_shape, gamma_scale, gamma_loc = self.distro_params[torf]
                color = 'g' if torf == 't' else 'r'
                count, bins, ignored = ax.hist(
                    self.data.score[self.truth if torf=='t' else ~self.truth], 'auto',
                    density=density, color=color, alpha=.5 if torf=='t' else .8
                )
                bins = bins - gamma_loc
                y = bins**(gamma_shape-1)*(np.exp(-bins/gamma_scale) /
                                           (sps.gamma(gamma_shape)*gamma_scale**gamma_shape))
                ax_y2.plot(bins, y, linewidth=2, color=color)
            
        elif self.distro == 'uniform':
            for torf in ('f','t'):
                unif_low, unif_high = self.distro_params[torf]
                color = 'g' if torf == 't' else 'r'
                count, bins, ignored = ax.hist(
                    self.data.score[self.truth if torf=='t' else ~self.truth], 'auto',
                    density=density, color=color, alpha=.5 if torf=='t' else .8
                )
                ax_y2.plot(bins, np.ones_like(bins), linewidth=2, color=color)

        return ax

    @staticmethod
    def empirical_param_calc(scores,trueGenes=None,distro='gamma',onlyPositiveScores=True):
        """Calculate distro param values

        Parameter value calculation based on real data.
        If no trueGenes are given, only calculates the average distribution.
        trueGenes should be bool pd.Series with same index as a pd.Series scores
        """
        if distro == 'gamma':
            if onlyPositiveScores: scores = scores.abs()
            if trueGenes is not None:
                fit_alpha_t, fit_loc_t, fit_beta_t = stats.gamma.fit(scores[trueGenes])
                fit_alpha_f, fit_loc_f, fit_beta_f = stats.gamma.fit(scores[~trueGenes])
                print('false gamma location was {} and should be close to 0!'.format(fit_loc_f))
                return {
                    'gamma_shape_t': fit_alpha_t,
                    'gamma_scale_t': fit_beta_t,
                    'gamma_loc_t': fit_loc_t,
                    'gamma_shape_f': fit_alpha_f,
                    'gamma_scale_f': fit_beta_f,                    
                }
            else:
                fit_alpha, fit_loc, fit_beta = stats.gamma.fit(scores)
                print('gamma location was {}. Should be close to 0!'.format(fit_loc))
                return {
                    'gamma_shape': fit_alpha,
                    'gamma_scale': fit_beta
                }



# -*- coding: utf-8 -*-
"""Benchmarking.

Module for comparing ninklings netwerk enrichment results
with other genset enrichment methods
"""
from . import Netwink
from .simulations import SimScores
from .config import config

class GenesetBenchmark:
    """Geneset Benchmark.

    Compares results of two methods.
    """
    def __init__(self,savedir,genesets,trueGenesets,netwink_kwargs={},simul_kwargs={}):
        """
        genesets and trueGenesets should be dict like,
        with geneset names as keys and geneset lists as values.
        """
        # Set base benchmark attributes
        self.genesets = genesets
        self.truth = trueGenesets
        self.trueGenes = {g for t in self.truth for g in self.truth[t]}
        # Load netwink
        netwink_defaults = {}
        netwink_defaults.update(netwink_kwargs)
        self.netwink = Netwink(savedir,**netwink_defaults)
        # Simulate scores
        self.sims = SimScores(
            self.netwink.get_nodes_series(),
            self.trueGenes,
            **simul_kwargs
        )
        # Prepare R data
        from bidali.retro import genesets2indices_r
        self.geneLabels = self.netwink.get_nodes_series().as_matrix()
        self.genesets_r = genesets2indices_r(self.genesets,self.geneLabels)

    def run_ni(self, kernel = 'random_walk', kernel_compute_kwargs = {}):
        """
        Run ninklings algorithm
        """
        self.niresults = self.netwink.gsea(
            self.genesets, self.sims.data.set_index('gene').score, kernel = kernel, minSize = 10,
            kernel_compute_kwargs = kernel_compute_kwargs
        )
    
    def run_opponent(self):
        """
        Runs the competing method
        """
        from bidali.retro import fgsea
        self.opresults = fgsea(self.genesets_r, self.sims.data.set_index('gene').score, minSize = 10)

    def compare_methods(self, permutations = 30, kernel = 'random_walk', simul_kwargs = {}, kernel_compute_kwargs = {}):
        """
        Runs run_ni and run_opponent x times (permutations)
        and compares results
        """
        import pandas as pd
        self.niresultsList = []
        self.opresultsList = []
        for i in range(permutations):
            self.sims = SimScores(
                self.netwink.get_nodes_series(),
                self.trueGenes,
                **simul_kwargs
            )
            print('Running ni {} ...'.format(i))
            self.run_ni(kernel = kernel, kernel_compute_kwargs = kernel_compute_kwargs)
            print('Running op {} ...'.format(i))
            self.run_opponent()
            self.niresultsList.append(self.niresults)
            self.opresultsList.append(self.opresults)
        # summarize comparison
        rankpositions = {}
        probabilities = {}
        for true_geneset in self.truth:
            rankpositions[(true_geneset,'ni')] = []
            rankpositions[(true_geneset,'op')] = []
            probabilities[(true_geneset,'ni')] = []
            probabilities[(true_geneset,'op')] = []
            for nir,opr in zip(self.niresultsList, self.opresultsList):
                rankpositions[(true_geneset,'ni')].append(nir.index.get_loc(true_geneset))
                rankpositions[(true_geneset,'op')].append(opr.set_index('pathway').index.get_loc(true_geneset))
                probabilities[(true_geneset,'ni')].append(nir.loc[true_geneset].prob)
                probabilities[(true_geneset,'op')].append(opr.set_index('pathway').loc[true_geneset].pval)
        self.rankcomparison = pd.DataFrame(rankpositions)
        self.probcomparison = pd.DataFrame(probabilities)

def main(args=None):
    """
    Runs benchmarking test for ninklings' algorithms
    """
    import os
    from bidali.LSD import get_msigdb6
    import numpy as np, pandas as pd
    import matplotlib.pyplot as plt
    mdb = get_msigdb6()
    trueGenesets = {
        'HALLMARK_E2F_TARGETS',
        'HALLMARK_G2M_CHECKPOINT'
    }
    bm = GenesetBenchmark(
        os.path.join(config['ninklings']['projectdir'],'ni_benchmark'),
        genesets = mdb['H'],
        trueGenesets = {k:mdb['H'][k] for k in mdb['H'].keys() & trueGenesets},
        netwink_kwargs = {'cosmicOnly': True}
    )
    restartProbs = np.arange(0,1.1,.1)
    kernel_kwargs_list = [
      {'restartProb':i} for i in restartProbs
    ]
    probcomparisons = []
    rankcomparisons = []
    for rrwkernel in kernel_kwargs_list:
        print('Restart probability:',rrwkernel['restartProb'])
        bm.compare_methods(kernel_compute_kwargs = rrwkernel, permutations = 10)
        probcomparisons.append(bm.probcomparison)
        rankcomparisons.append(bm.rankcomparison)
    # Plot summaries
    means_probcomparisons = pd.DataFrame(pd.concat([df.mean() for df in probcomparisons]),columns=['probMean'])
    means_probcomparisons['restartProb'] = np.repeat(restartProbs, 4)
    means_probcomparisons['probVar'] = pd.concat([df.var() for df in probcomparisons])
    fig,ax1=plt.subplots()
    for gs in trueGenesets:
        for method in ('ni','op'):
            ax1.errorbar(
                means_probcomparisons.loc[gs,method].restartProb,
                means_probcomparisons.loc[gs,method].probMean,
                yerr = means_probcomparisons.loc[gs,method].probVar ** .5,
                fmt='o',
                label = '{} - {}'.format(gs,method)
            )
    ax1.legend()
    ax1.set_title('Probability comparison')
    ax1.set_xlabel('Random restart probability')
    means_rankcomparisons = pd.DataFrame(pd.concat([df.mean() for df in rankcomparisons]),columns=['rankMean'])
    means_rankcomparisons['restartProb'] = np.repeat(restartProbs, 4)
    means_rankcomparisons['rankVar'] = pd.concat([df.var() for df in rankcomparisons])
    fig,ax2=plt.subplots()
    for gs in trueGenesets:
        for method in ('ni','op'):
            ax2.errorbar(
                means_rankcomparisons.loc[gs,method].restartProb,
                means_rankcomparisons.loc[gs,method].rankMean,
                yerr = means_rankcomparisons.loc[gs,method].rankVar ** .5,
                fmt='o',
                label = '{} - {}'.format(gs,method)
            )
    ax2.legend()
    ax2.set_title('Rank comparison')
    ax2.set_xlabel('Random restart probability')
    bm.rankcomparisons = means_rankcomparisons
    bm.probcomparisons = means_probcomparisons
    return (bm, ax1, ax2)




# -*- coding: utf-8 -*-
"""Kernels.

Kernel base (interface) class and implementing classes.
"""

import random, numpy as np
import pandas as pd
import matplotlib.pyplot as plt

# Kernels
class Kernel():
    def __init__(self,netwink,kernelspec={}):
        """
        netwink: the network object with which the kernel is associated
        kernelspec: dict with parameters for kernel computation
        """
        self.netwink = netwink
        self.spec = kernelspec

    def __repr__(self):
        if 'computedMatrix' in vars(self):
            return self.computedMatrix.__repr__()

    def __str__(self):
        return '{}_{}_{}'.format(
            self.netwink.name,
            self.__class__.__name__,
            '_'.join(['{}_{}'.format(k,self.spec[k]) for k in sorted(self.spec)]) if self.spec else 'default'
        )

    def __hash__(self):
        return hash(str(self))

    def compute(self):
        """
        sets self.computedMatrix
        returns self
        """
        raise NotImplementedError('implement in inheriting classes')

    def score_geneset(self,geneset,method='vector',ax=None,c='r'):
        """
        Calculate the sum of a geneset.
        netwink.apply_gene_scores(genescores) has to been run before
        
        The function spreads the impact of the genes in the geneset of interest
        by constructing a vector as big as the network (1 for gene of interest, 0 otherwise),
        elementwise multpication with kernel computed matrix
        sum defines the subnetwork, but permutation test is needed to test how relevant
        its sum is, compared to scores randomly same-sized genesets get.
        """
        if method == 'vector':
            diffusion =  self.netwink.scoresVector @ self.computedMatrix
            score = diffusion.T[self.netwink.get_nodes_series().isin(geneset)].sum()
        elif method == 'matrix':
            score = np.multiply(
                self.computedMatrix,
                np.multiply(self.netwink.scoresMatrix,self.netwink.subset_adjmatrix(geneset))
            ).sum()
        if ax: ax.axvline(score,c=c)
        return score

    def permutate_geneset_scores(self,geneset,numberOfPermutations=1000,seed=None,ax=None):
        """
        Takes the length of the passed geneset, then makes `numberOfPermutations` genesets
        of the same length and calculates their scores. Returns a list with the scores.

        seed can be used to (re)set the random state
        """
        genesetLength = len(geneset)
        allGenes = set(self.netwink.get_nodes_series())
        if seed: random.seed(seed)
        shelveKey = 'Geneset length {} - permutations {} - seed {} - kernel hash {}'.format(
            genesetLength,numberOfPermutations,seed,hash(self)
            #optionally also include a hash of the netwink nodes+edges to be sure it did not change
        )
        try: scores = self.netwink.shelve[shelveKey]
        except KeyError:
            scores = [
                self.score_geneset(random.sample(allGenes,genesetLength))
                for i in range(numberOfPermutations)
            ]
            self.netwink.shelve[shelveKey] = scores
            #logger.
        if ax:
            sns.distplot(scores,ax=ax)
        return scores
        
    def visualize(self):
        plt.matshow(self.computedMatrix)

    def convert_to_distance(self):
        """
        convert a kernel computed matrix to a distance matrix by subtrating from 1
        and equaling the diagonale positions to 0 distance
        
        the distance matrix can be used for clustering related genes
        this is good to use in a check, e.g. checking enrichment of biological functions
        """
        # Take average values in case computedMatrix is not symmetrical
        averageMatrix = (self.computedMatrix + self.computedMatrix.T)/2
        # Substract from 1 to get distance instead of connectedness and set diagonal distances to 0
        distanceMatrix = np.multiply(
            1 - averageMatrix,
            np.logical_not(np.matrix(np.diag(np.ones(len(averageMatrix)))))
        )
        return distanceMatrix
        
## laplacian exponential diffusion kernel
class ExponentialDiffusionKernel(Kernel):
    def compute(self):
        """
        Spec parameters:
        alpha (default: 1)
        """
        # Set parameters
        alpha = 1 if 'alpha' not in self.spec else self.spec['alpha']
        # Compute
        from scipy.linalg import expm
        self.degreematrix = np.diag(np.array(self.netwink.admatrix.sum(axis=1)).flatten())
        self.laplacian = self.degreematrix - self.netwink.admatrix
        self.computedMatrix = expm(-alpha*np.array(self.laplacian)) # => wel een kernel
        return self

## restart random walk
class RestartRandomWalk(Kernel):
    def compute(self):
        """
        computed restartmatrix is not symmetrical

        Spec parameters:
        restartProb (default: 0.1)
        """
        # Set parameters
        restartProb = 0.1 if 'restartProb' not in self.spec else self.spec['restartProb']
        # Compute
        self.degreematrix = np.diag(np.array(self.netwink.admatrix.sum(axis=1)).flatten())
        self.inversematrix = np.linalg.inv(self.degreematrix - ((1-restartProb)*self.netwink.admatrix))
        self.restartmatrix = self.inversematrix*self.degreematrix # => op factor na probabiliteiten => geen kernel
        self.computedMatrix = self.restartmatrix
        return self



from setuptools import setup

package = "pyni"
version = "0.0.1"

setup(name = package,
      version = version,
      description="Python network inference package",
      url='https://github.ugent.be/cvneste/pyni',
      author = 'Christophe Van Neste',
      author_email = 'christophe.vanneste@ugent.be',
      license = 'GNU GENERAL PUBLIC LICENSE',
      packages = ['pyni'],
      install_requires = [
          'matplotlib',
          'pandas',
          'scipy',
          'requests',
          'networkx',
          'pydot',
          'bidali'
      ],
      extras_require = {
          'documentation': ['Sphinx']
      },
      package_data = {
          'pyni': [
          ]
      },
      include_package_data = False,
      zip_safe = False,
      entry_points = {
          'console_scripts': ['ni=pyni.__main__:main'],
      },
      test_suite = 'nose.collector',
      tests_require = ['nose']
)

#To install with symlink, so that changes are immediately available:
#pip install -e .



#!/usr/bin/env python3
import networkx as nx
import numpy as np
from scipy.linalg import expm
from collections import OrderedDict
import pandas as pd
from io import  StringIO
import os, logging, random, shelve
import matplotlib.pyplot as plt
import seaborn as sns
from bidali.LSD.dealer.external.ensembl import get_biomart

# Set matplotlib to interactive mode when executed from interactive shell
if 'ps1' in vars(os.sys): plt.ion()

# Configuration
from .config import config

# Kernels
from .kernels import ExponentialDiffusionKernel, RestartRandomWalk

class Netwink():
    """
    Ninkling's starting class.
    Does the setup for all down stream network analyses
    """

    def __init__(self,savedir,name=None,loglevel='DEBUG',cosmicOnly=True,fullInit=True):
        self.location = savedir
        self.name = name if name else os.path.basename(savedir)
        self.logger = logging.Logger(self.name,level=loglevel)
        if fullInit:
            self.check_location()
            self.setup_logging(loglevel)
            self.shelve = shelve.open(os.path.join(self.location,'distrostore'))
            self.set_annotation()
            self.load_network(cosmicOnly)
            self.associated_kernels = {}
            self.logger.debug('%s fully created',self)
        else: self.logger.warning('%s created but no network loaded',self)

    def __del__(self):
        self.shelve.close()

    def __repr__(self):
        return '<Netwink object: {}>'.format(self.name)

    def check_location(self):
        if not os.path.exists(self.location):
            os.mkdir(self.location)

    def setup_logging(self,level='DEBUG'):
        stdout = logging.StreamHandler(stream=os.sys.stdout)
        stdout.setLevel(level)
        logfile = logging.FileHandler(os.path.join(self.location,self.name+'.log'))
        logfile.setLevel(level)
        self.logger.addHandler(stdout)
        self.logger.addHandler(logfile)
        
    def set_annotation(self):
        self.annotationFile = os.path.join(self.location, 'ensembl_annotation.tsv')
        if os.path.exists(self.annotationFile):
            self.annotation = pd.read_table(self.annotationFile)
        else:
            self.annotation = get_biomart()
            self.annotation = self.annotation[
                self.annotation['Source of gene name']=='HGNC Symbol'
            ]
            self.annotation = self.annotation[
                ~self.annotation['Gene name'].duplicated()
            ] #TODO ask server to not send duplicate ids
            self.annotation['network_id'] = self.annotation['Gene stable ID'].apply(lambda x: int(x.replace('ENSG','')))
            self.annotation.to_csv(self.annotationFile,sep='\t')

    def load_network(self,cosmicOnly=False):
        # if complexes in network -> condir complex as its own entity?
        # for brute force analysis best not
        # TODO how are we going to deal with it?
        self.networkFile = os.path.join(self.location, 'reactome_FI_filteredEdges.tsv')
        if os.path.exists(self.networkFile):
            self.networkgenes = pd.read_table(self.networkFile)
        else:
            networktable = pd.read_table(os.path.join(config['ninklings']['datadir'],'reactome_FI.txt'))
            networktable = networktable[networktable.Gene1.isin(self.annotation['Gene name'])]
            networktable = networktable[networktable.Gene2.isin(self.annotation['Gene name'])]
            networktable = networktable[networktable.Score >= .75]
            print('unique genes:',len(set(networktable.Gene1)|set(networktable.Gene2)))
            self.networkgenes = networktable[['Gene1','Gene2']]
            self.networkgenes.to_csv(self.networkFile, sep='\t')
        # Construct network
        self.graph = nx.Graph()
        self.networkgenes.T.apply(lambda x: self.graph.add_edge(x['Gene1'],x['Gene2']))
        if cosmicOnly:
            cosmicgenes = set(pd.read_table(
                os.path.join(config['ninklings']['datadir'],'cosmic_20180125.tsv')
            )['Gene Symbol'])
            self.graph = self.graph.subgraph(cosmicgenes)
            components = {
                comp.number_of_nodes():comp
                for comp in nx.connected_component_subgraphs(self.graph)
            }
            self.graph = components[max(components)]
        self.admatrix = nx.adjacency_matrix(self.graph).todense()
        self.admatrix = self.admatrix.astype(np.float32)
        # Set weights to default of 1
        self.weights = np.ones(self.admatrix.shape)

    def load_weights(self,weights):
        """
        Load weights such as context relevant correlation values

        TODO what to do with NA weights?
        """
        self.weights = weights

    def load_correlations(self,nodesCorrelationData,method='pearson',expSmoothening=1,**kwargs):
        """
        Calculate correlations based on the data in nodesCorrelationData pd.DataFrame
        and load them as weights

        if nodesCorrelationData is not pd.DataFrame, it is assumed to be a filename,
        that will be used to load the DataFrame with pd.read_csv and kwargs provided.

        correlation method => any method accepted by DataFrame.corr (pearson, kendall, spearman)

        expSmoothening => all corr values are transformed to the power of expSmoothening
          values furhter away from 1 will go exponentially towards 0 with higher expSmoothening value
        """
        if type(nodesCorrelationData) != pd.DataFrame:
            nodesCorrelationData = pd.read_csv(nodesCorrelationData,**kwargs)
        # Filter non-node data
        nodesCorrelationData = nodesCorrelationData.reindex(self.get_nodes_series())
        correlations = nodesCorrelationData.T.corr(method=method)
        self.correlationSigns = np.sign(correlations)
        weights = correlations.abs() ** expSmoothening
        self.load_weights(weights)
        

    def get_nodes_series(self):
        return pd.Series(self.graph.nodes)

    def export_admatrix(self):
        self.admatrix.tofile(os.path.join(self.location,'admatrix.tsv'), sep='\t')

    def plot_admatrix(self):
        plt.matshow(self.admatrix)

    def apply_gene_scores(self,genescores,fillna=0):
        """
        Apply gene scores to the adjacency matrix. 
        Genescores should be dict with as keys the node names.
        When a node/gene is not in genescores it gets a score of `fillna`

        In the returned matrix, each gene connected to the row's gene
        gets the score of that gene. The returned matrix is not symmetrical.
        """
        self.scoresVector = self.get_nodes_series().map(genescores).fillna(fillna).astype(np.float32).as_matrix()
        scoresDiag = np.matrix(np.diag(self.scoresVector))
        self.scoresMatrix = scoresDiag * np.matrix(self.admatrix)
        
    
    def subset_adjmatrix(self,geneset,externalEdges=False):
        """
        For a given set of genes, returns the original adjancency matrix
        with 1 indicating edges within the geneset, or if externalEdges is
        True, edges between 1 of the genes of interest and any other gene.
        Genes that are not in the geneset or not connected to a geneset gene
        if externalEdges is True are 0.
        """
        return np.multiply(
            self.admatrix,
            self.get_geneset_selection_matrix(geneset,externalEdges)
        )

    def get_geneset_selection_matrix(self,geneset,externalEdges=False):
        """
        For a given set of genes, returns a matrix with row and columns
        1 for genes of the set, and 0 for all others if externalEdges is True.
        If externalEdges is False only 1 for combinations of genes within
        the geneset, and 0 for all others.
        """
        genesetVector = self.get_nodes_series().isin(geneset).astype(np.float32).as_matrix()
        if externalEdges:
            return np.logical_or(
                np.outer(genesetVector,np.ones(len(genesetVector))),
                np.outer(np.ones(len(genesetVector)),genesetVector)
            ).astype(np.float32)
        else:
            return np.outer(genesetVector,genesetVector)

    def visualize(self,outfiletype='pdf'):
        """Visualize graph

        Uses the Graphviz program
        """
        import pydot
        from networkx.drawing.nx_pydot import write_dot
        from plumbum import local
        filename = os.path.join(self.location,'grid.dot')
        write_dot(self.graph, filename)
        local['neato']('-T'+outfiletype,filename,'-o{}{}'.format(filename[:-3],outfiletype))        

    def gsea(self,genesets, scores, kernel, nperm = 10000, minSize = 10, kernel_compute_kwargs = {}):
        """
        Calculate a score and p-value for every geneset in genesets
        with gene scores and a defined kernel

        kernel options: 'random_walk', 'exp'

        nperm => number of permutations

        minsize a genesets needs after filtering genes not in the analysed network
        """
        kernels = {
            'random_walk': RestartRandomWalk,
            'exp': ExponentialDiffusionKernel
        }
        kernel = kernels[kernel](self, kernelspec = kernel_compute_kwargs)
        if str(kernel) in self.associated_kernels:
            # if kernel has been computed earlier, swap for computed kernel
            kernel = self.associated_kernels[str(kernel)]
        else: self.associated_kernels[str(kernel)] = kernel.compute()
        # apply scores
        self.apply_gene_scores(scores)
        # prune genesets
        geneset_originalSizes = {gs:len(genesets[gs]) for gs in genesets}
        allGenes = set(self.get_nodes_series())
        genesets = {gs:[g for g in genesets[gs] if g in allGenes] for gs in genesets}
        geneset_sizes = {gs:len(genesets[gs]) for gs in genesets}
        # calculate scores
        geneset_scores = {}
        geneset_probs = {}
        for gs in genesets:
            geneset_scores[gs] = kernel.score_geneset(genesets[gs])
            geneset_probs[gs] = (pd.Series(
                kernel.permutate_geneset_scores(genesets[gs],numberOfPermutations=nperm)
            ) > geneset_scores[gs]).mean()
        results = pd.DataFrame(
            {'score': geneset_scores, 'prob': geneset_probs, 'size': geneset_sizes, 'osize': geneset_originalSizes},
            columns = ['osize','size','score','prob']
        ).sort_values('prob')
        results = results[results['size'] >= minSize]
        return results
    
